@@ -31,20 +31,28 @@
 #include <QtCore/QXmlStreamReader>
 #include <QtCore/QFile>
 #include <QtCore/QDir>
+#include <fileout.h>
+#include <limits>
 
 EXPORT_GENERATOR_PLUGIN(new QtDocGenerator)
 
 static Indentor INDENT;
 
-namespace
+static bool shouldSkip(const AbstractMetaFunction* func)
 {
+    return func->isConstructor()
+           || func->isModifiedRemoved()
+           || func->declaringClass() != func->ownerClass()
+           || func->isCastOperator()
+           || func->name() == "operator=";
+}
 
 static bool functionSort(const AbstractMetaFunction *func1, const AbstractMetaFunction *func2)
 {
     return func1->name() < func2->name();
 }
 
-QString createRepeatedChar(int i, char c)
+static QString createRepeatedChar(int i, char c)
 {
     QString out;
     for (int j = 0; j < i; ++j)
@@ -53,21 +61,19 @@ QString createRepeatedChar(int i, char c)
     return out;
 }
 
-QString escape(QString& str)
+static QString escape(QString& str)
 {
     return str
             .replace("*", "\\*")
             .replace("_", "\\_");
 }
 
-QString escape(const QStringRef& strref)
+static QString escape(const QStringRef& strref)
 {
     QString str = strref.toString();
     return escape(str);
 }
 
-}
-
 QtXmlToSphinx::QtXmlToSphinx(QtDocGenerator* generator, const QString& doc, const QString& context)
         : m_context(context), m_generator(generator), m_insideBold(false), m_insideItalic(false)
 {
@@ -95,6 +101,7 @@ QtXmlToSphinx::QtXmlToSphinx(QtDocGenerator* generator, const QString& doc, cons
     m_handlerMap.insert("underline", &QtXmlToSphinx::handleItalicTag);
     m_handlerMap.insert("superscript", &QtXmlToSphinx::handleSuperScriptTag);
     m_handlerMap.insert("code", &QtXmlToSphinx::handleCodeTag);
+    m_handlerMap.insert("badcode", &QtXmlToSphinx::handleCodeTag);
     m_handlerMap.insert("legalese", &QtXmlToSphinx::handleCodeTag);
     m_handlerMap.insert("section", &QtXmlToSphinx::handleAnchorTag);
     m_handlerMap.insert("quotefile", &QtXmlToSphinx::handleQuoteFileTag);
@@ -207,12 +214,31 @@ QString QtXmlToSphinx::transform(const QString& doc)
     return retval;
 }
 
-QString QtXmlToSphinx::readFromLocation(QString& location, QString& identifier)
+QString QtXmlToSphinx::readFromLocations(const QStringList& locations, const QString& path, const QString& identifier)
+{
+    QString result;
+    bool ok;
+    foreach (QString location, locations) {
+        location.append('/');
+        location.append(path);
+        result = readFromLocation(location, identifier, &ok);
+        if (ok)
+            break;
+    }
+    if (!ok)
+        ReportHandler::warning("Couldn't read code snippet file: {"+ locations.join("|") + '}' + path);
+    return result;
+}
+
+QString QtXmlToSphinx::readFromLocation(const QString& location, const QString& identifier, bool* ok)
 {
     QFile inputFile;
     inputFile.setFileName(location);
     if (!inputFile.open(QIODevice::ReadOnly)) {
-        ReportHandler::warning("Couldn't read code snippet file: "+inputFile.fileName());
+        if (!ok)
+            ReportHandler::warning("Couldn't read code snippet file: "+inputFile.fileName());
+        else
+            *ok = false;
         return QString();
     }
 
@@ -237,9 +263,11 @@ QString QtXmlToSphinx::readFromLocation(QString& location, QString& identifier)
         }
     }
 
-    if (code.isEmpty())
+    if (!identifierIsEmpty && !getCode)
         ReportHandler::warning("Code snippet file found ("+location+"), but snippet "+ identifier +" not found.");
 
+    if (ok)
+        *ok = true;
     return code;
 }
 
@@ -339,8 +367,7 @@ void QtXmlToSphinx::handleSnippetTag(QXmlStreamReader& reader)
         }
         QString location = reader.attributes().value("location").toString();
         QString identifier = reader.attributes().value("identifier").toString();
-        location.prepend(m_generator->codeSnippetDir() + '/');
-        QString code = readFromLocation(location, identifier);
+        QString code = readFromLocations(m_generator->codeSnippetDirs(), location, identifier);
         if (!consecutiveSnippet)
             m_output << INDENT << "::\n\n";
 
@@ -522,7 +549,8 @@ void QtXmlToSphinx::handleLinkTag(QXmlStreamReader& reader)
             QStringList rawlinklist = l_linkref.split(".");
             if (rawlinklist.size() == 1 || rawlinklist.first() == m_context) {
                 QString context = resolveContextForMethod(rawlinklist.last());
-                l_linkref.prepend(context + '.');
+                if (!l_linkref.startsWith(context))
+                    l_linkref.prepend(context + '.');
             }
         } else if (l_type == "function" && m_context.isEmpty()) {
             l_linktag = " :func:`";
@@ -804,10 +832,6 @@ QTextStream& operator<<(QTextStream& s, const QtXmlToSphinx::Table &table)
     return s;
 }
 
-static QString getClassName(const AbstractMetaClass *cppClass) {
-    return QString(cppClass->typeEntry()->qualifiedCppName()).replace("::", ".");
-}
-
 static QString getFuncName(const AbstractMetaFunction *cppFunc) {
     static bool hashInitialized = false;
     static QHash<QString, QString> operatorsHash;
@@ -846,9 +870,18 @@ static QString getFuncName(const AbstractMetaFunction *cppFunc) {
     return result.replace("::", ".");
 }
 
+QtDocGenerator::QtDocGenerator() : m_docParser(new QtDocParser)
+{
+}
+
+QtDocGenerator::~QtDocGenerator()
+{
+    delete m_docParser;
+}
+
 QString QtDocGenerator::fileNameForClass(const AbstractMetaClass *cppClass) const
 {
-    return QString("%1.rst").arg(getClassName(cppClass));
+    return QString("%1.rst").arg(getClassTargetFullName(cppClass, false));
 }
 
 void QtDocGenerator::writeFormatedText(QTextStream& s, const Documentation& doc, const AbstractMetaClass* metaClass)
@@ -856,39 +889,39 @@ void QtDocGenerator::writeFormatedText(QTextStream& s, const Documentation& doc,
     QString metaClassName;
 
     if (metaClass)
-        metaClassName = getClassName(metaClass);
+        metaClassName = getClassTargetFullName(metaClass);
 
     if (doc.format() == Documentation::Native) {
         QtXmlToSphinx x(this, doc.value(), metaClassName);
         s << x;
     } else {
-        s << doc.value();
+        QStringList lines = doc.value().split("\n");
+        QRegExp regex("\\S"); // non-space character
+        int typesystemIndentation = std::numeric_limits<int>().max();
+        // check how many spaces must be removed from the begining of each line
+        foreach (QString line, lines) {
+            int idx = line.indexOf(regex);
+            if (idx >= 0)
+                typesystemIndentation = qMin(typesystemIndentation, idx);
+        }
+        foreach (QString line, lines)
+            s << INDENT << line.remove(0, typesystemIndentation) << endl;
     }
 
     s << endl;
 }
 
-void QtDocGenerator::writeFunctionBrief(QTextStream &s,
-                                      const AbstractMetaClass *cppClass,
-                                      const AbstractMetaFunction *cppFunction)
+void QtDocGenerator::generateClass(QTextStream &s, const AbstractMetaClass *metaClass)
 {
-    s << INDENT << "def :meth:`"
-      << cppFunction->name() << "<";
-    if (cppClass && cppClass->name() != cppFunction->name())
-        s << getClassName(cppClass) << '.';
+    ReportHandler::debugSparse("Generating Documentation for " + metaClass->fullName());
 
-    s << cppFunction->name() << ">`"
-      << " (" << parseArgDocStyle(cppClass, cppFunction) << "):";
-}
+    m_packages[metaClass->package()] << fileNameForClass(metaClass);
 
-void QtDocGenerator::generateClass(QTextStream &s, const AbstractMetaClass *cppClass)
-{
-    QString doc;
-    QTextStream  doc_s(&doc);
+    m_docParser->setPackageName(metaClass->package());
+    m_docParser->fillDocumentation(const_cast<AbstractMetaClass*>(metaClass));
 
-    ReportHandler::debugSparse("Generating Documentation for " + cppClass->fullName());
-    s << ".. module:: " << packageName() << endl;
-    QString className = getClassName(cppClass);
+    s << ".. module:: " << metaClass->package() << endl;
+    QString className = getClassTargetFullName(metaClass, false);
     s << ".. _" << className << ":" << endl << endl;
 
     s << className << endl;
@@ -897,102 +930,73 @@ void QtDocGenerator::generateClass(QTextStream &s, const AbstractMetaClass *cppC
     s << ".. inheritance-diagram:: " << className << endl
       << "    :parts: 2" << endl << endl; // TODO: This would be a parameter in the future...
 
+    writeFunctionList(s, metaClass);
+
     //Function list
-    AbstractMetaFunctionList functionList = filterFunctions(cppClass);
+    AbstractMetaFunctionList functionList = metaClass->functions();
     qSort(functionList.begin(), functionList.end(), functionSort);
 
-    doc_s << "Detailed Description\n"
-             "--------------------\n\n";
+    s << "Detailed Description\n"
+         "--------------------\n\n";
 
-    writeInjectDocumentation(doc_s, DocModification::Prepend, cppClass, 0);
-    writeFormatedText(doc_s, cppClass->documentation(), cppClass);
+    writeInjectDocumentation(s, DocModification::Prepend, metaClass, 0);
+    if (!writeInjectDocumentation(s, DocModification::Replace, metaClass, 0))
+        writeFormatedText(s, metaClass->documentation(), metaClass);
+
+    if (!metaClass->isNamespace())
+        writeConstructors(s, metaClass);
+    writeEnums(s, metaClass);
+    if (!metaClass->isNamespace())
+        writeFields(s, metaClass);
 
 
-    if (!cppClass->isNamespace()) {
+    foreach (AbstractMetaFunction* func, functionList) {
+        if (shouldSkip(func))
+            continue;
 
-        writeConstructors(doc_s, cppClass);
-        writeEnums(doc_s, cppClass);
-        writeFields(doc_s, cppClass);
+        if (func->isStatic())
+            s <<  ".. staticmethod:: ";
+        else
+            s <<  ".. method:: ";
 
-        foreach (AbstractMetaFunction *func, functionList) {
-            if ((func->isConstructor() || func->isModifiedRemoved()) ||
-                (func->declaringClass() != cppClass))
-                continue;
-
-            if (func->isStatic())
-                doc_s <<  ".. staticmethod:: ";
-            else
-                doc_s <<  ".. method:: ";
-
-            writeFunction(doc_s, true, cppClass, func);
-        }
+        writeFunction(s, true, metaClass, func);
     }
 
-    writeInjectDocumentation(doc_s, DocModification::Append, cppClass, 0);
-
-    writeFunctionList(s, doc, cppClass);
-
-    s << doc;
+    writeInjectDocumentation(s, DocModification::Append, metaClass, 0);
 }
 
-QString QtDocGenerator::parseFunctionDeclaration(const QString &doc, const AbstractMetaClass *cppClass)
-{
-    //.. method:: QObject.childEvent(arg__1)
-    //def :meth:`removeEventFilter<QObject.removeEventFilter>` (arg__1):
-
-    QString data = doc;
-    QString markup;
-
-    if (data.startsWith(".. method::"))
-        markup = ".. method::";
-    else if (data.startsWith(".. staticmethod::"))
-        markup = ".. staticmethod::";
-    else
-        return QString();
-
-    data = data.mid(markup.size()); //remove .. method::
-    data = data.mid(data.indexOf(".") + 1); //remove class name
-
-    QString methName = data.mid(0, data.indexOf("("));
-    QString methArgs = data.mid(data.indexOf("("));
-
-    QString scope = cppClass->name();
-    QStringList splittedMethName = methName.split(".");
-
-    if (splittedMethName.first() == scope) {
-        splittedMethName.removeFirst();
-        methName = splittedMethName.join(".");
-    }
-    scope.append(".");
-
-    data = QString("def :meth:`%1<%2%3>` %4")
-        .arg(methName)
-        .arg(scope)
-        .arg(methName)
-        .arg(methArgs);
-
-    return data;
-}
-
-
-void QtDocGenerator::writeFunctionList(QTextStream &s, const QString &content, const AbstractMetaClass *cppClass)
+void QtDocGenerator::writeFunctionList(QTextStream& s, const AbstractMetaClass* cppClass)
 {
     QStringList functionList;
     QStringList staticFunctionList;
 
-    QStringList lst = content.split("\n");
-    foreach(QString row, lst) {
-        QString data = row.trimmed();
-        if (data.startsWith(".. method::")) {
-            functionList << parseFunctionDeclaration(data, cppClass);
-        }
-        else if (data.startsWith(".. staticmethod::")) {
-            staticFunctionList << parseFunctionDeclaration(data, cppClass);
-        }
+    foreach (AbstractMetaFunction* func, cppClass->functions()) {
+        if (shouldSkip(func))
+            continue;
+
+        QString className;
+        if (!func->isConstructor())
+            className =  getClassTargetFullName(cppClass) + '.';
+        else if (func->implementingClass() && func->implementingClass()->enclosingClass())
+            className =  getClassTargetFullName(func->implementingClass()->enclosingClass()) + '.';
+        QString funcName = getFuncName(func);
+
+        QStringList& list = func->isStatic() ? staticFunctionList : functionList;
+        QString str("def :meth:`");
+
+        str += funcName;
+        str += '<';
+        if (!funcName.startsWith(className))
+            str += className;
+        str += funcName;
+        str += ">` (";
+        str += parseArgDocStyle(cppClass, func);
+        str += ')';
+
+        list << str;
     }
 
-    if ((functionList.size() > 0) || (staticFunctionList.size() > 0))
-    {
+    if ((functionList.size() > 0) || (staticFunctionList.size() > 0)) {
         QtXmlToSphinx::Table functionTable;
         QtXmlToSphinx::TableRow row;
 
@@ -1037,7 +1041,7 @@ void QtDocGenerator::writeEnums(QTextStream& s, const AbstractMetaClass* cppClas
     static const QString section_title(".. attribute:: ");
 
     foreach (AbstractMetaEnum *en, cppClass->enums()) {
-        s << section_title << getClassName(cppClass) << "." << en->name() << endl << endl;
+        s << section_title << getClassTargetFullName(cppClass) << "." << en->name() << endl << endl;
         writeFormatedText(s, en->documentation(), cppClass);
     }
 }
@@ -1047,7 +1051,7 @@ void QtDocGenerator::writeFields(QTextStream &s, const AbstractMetaClass *cppCla
     static const QString section_title(".. attribute:: ");
 
     foreach (AbstractMetaField *field, cppClass->fields()) {
-        s << section_title << getClassName(cppClass) << "." << field->name() << endl << endl;
+        s << section_title << getClassTargetFullName(cppClass) << "." << field->name() << endl << endl;
         //TODO: request for member ‘documentation’ is ambiguous
         writeFormatedText(s, field->AbstractMetaAttributes::documentation(), cppClass);
     }
@@ -1076,8 +1080,8 @@ void QtDocGenerator::writeConstructors(QTextStream &s, const AbstractMetaClass *
         writeFunction(s, false, cppClass, func);
         foreach(AbstractMetaArgument *arg, func->arguments())
         {
-            if (!arg_map.contains(arg->argumentName())) {
-                arg_map.insert(arg->argumentName(), arg);
+            if (!arg_map.contains(arg->name())) {
+                arg_map.insert(arg->name(), arg);
             }
         }
     }
@@ -1099,30 +1103,42 @@ void QtDocGenerator::writeConstructors(QTextStream &s, const AbstractMetaClass *
 QString QtDocGenerator::parseArgDocStyle(const AbstractMetaClass *cppClass, const AbstractMetaFunction *func)
 {
     QString ret;
-    bool optional = false;
+    int optArgs = 0;
 
-    foreach (AbstractMetaArgument *arg, func->arguments()) {
+    foreach (AbstractMetaArgument* arg, func->arguments()) {
 
         if (func->argumentRemoved(arg->argumentIndex() + 1))
             continue;
 
-        if (arg->argumentIndex() > 0)
-            ret += ",";
-
-        if (!arg->defaultValueExpression().isEmpty() && (!optional)) {
-            ret += "[";
-            optional = true;
+        bool thisIsoptional = !arg->defaultValueExpression().isEmpty();
+        if (optArgs || thisIsoptional) {
+            ret += '[';
+            optArgs++;
         }
 
-        ret += arg->argumentName();
+        if (arg->argumentIndex() > 0)
+            ret += ", ";
 
-        if (optional)
-            ret += "=" + arg->defaultValueExpression();
+        ret += arg->name();
+
+        if (thisIsoptional) {
+            QString defValue = arg->defaultValueExpression();
+            if (defValue == "QString()") {
+                defValue = "\"\"";
+            } else if (defValue == "QStringList()" || defValue.startsWith("QVector") || defValue.startsWith("QList")) {
+                defValue = "list()";
+            } else if (defValue == "QVariant()") {
+                defValue = "None";
+            } else {
+                defValue.replace("::", ".");
+                if (defValue == "0" && (arg->type()->isQObject() || arg->type()->isObject()))
+                    defValue = "None";
+            }
+            ret += "=" + defValue;
+        }
     }
 
-    if (optional)
-        ret += "]";
-
+    ret += QString(']').repeated(optArgs);
     return ret;
 }
 
@@ -1189,12 +1205,13 @@ void QtDocGenerator::writeDocSnips(QTextStream &s,
     }
 }
 
-void QtDocGenerator::writeInjectDocumentation(QTextStream &s,
+bool QtDocGenerator::writeInjectDocumentation(QTextStream &s,
                                             DocModification::Mode mode,
                                             const AbstractMetaClass *cppClass,
                                             const AbstractMetaFunction *func)
 {
     Indentation indentation(INDENT);
+    bool didSomething = false;
 
     foreach (DocModification mod, cppClass->typeEntry()->docModifications()) {
         if (mod.mode() == mode) {
@@ -1212,16 +1229,18 @@ void QtDocGenerator::writeInjectDocumentation(QTextStream &s,
                     continue;
 
                 doc.setValue(mod.code() , fmt);
-                s << INDENT;
                 writeFormatedText(s, doc, cppClass);
+                didSomething = true;
             }
         }
     }
 
     s << endl;
 
+    // TODO: Deprecate the use of doc string on glue code.
+    //       This is pre "add-function" and "inject-documentation" tags.
     if (func) {
-        writeDocSnips(s, getCodeSnips(func),
+        writeDocSnips(s, func->injectedCodeSnips(),
                        (mode == DocModification::Prepend ? CodeSnip::Beginning : CodeSnip::End),
                        TypeSystem::TargetLangCode);
     } else {
@@ -1229,53 +1248,69 @@ void QtDocGenerator::writeInjectDocumentation(QTextStream &s,
                        (mode == DocModification::Prepend ? CodeSnip::Beginning : CodeSnip::End),
                        TypeSystem::TargetLangCode);
     }
+    return didSomething;
 }
 
 void QtDocGenerator::writeFunctionSignature(QTextStream& s, const AbstractMetaClass* cppClass, const AbstractMetaFunction* func)
 {
+    QString className;
     if (!func->isConstructor())
-        s << getClassName(cppClass) << '.';
+        className =  getClassTargetFullName(cppClass) + '.';
     else if (func->implementingClass() && func->implementingClass()->enclosingClass())
-        s << func->implementingClass()->enclosingClass()->name() << '.';
-    s << getFuncName(func) << "(" << parseArgDocStyle(cppClass, func) << ")";
+        className =  getClassTargetFullName(func->implementingClass()->enclosingClass()) + '.';
+
+    QString funcName = getFuncName(func);
+    if (!funcName.startsWith(className))
+        funcName = className + funcName;
+
+    s << funcName << "(" << parseArgDocStyle(cppClass, func) << ")";
 }
 
 QString QtDocGenerator::translateToPythonType(const AbstractMetaType *type, const AbstractMetaClass *cppClass)
 {
-    QString originalType = translateType(type, cppClass, Options(ExcludeConst) | ExcludeReference);
-    QString strType = originalType;
-
-    //remove "*"
-    strType.remove("*");
-    TypeEntry *te = TypeDatabase::instance()->findType(originalType.trimmed());
-    if (te) {
-        return te->targetLangName();
-    } else {
-        //remove <, >
+    QString strType;
+    if (type->name() == "QString") {
+        strType = "unicode";
+    } else if (type->name() == "QVariant") {
+        strType = "object";
+    } else if (type->name() == "QStringList") {
+        strType = "list of strings";
+    } else if (type->isConstant() && type->name() == "char" && type->indirections() == 1) {
+        strType = "str";
+    } else if (type->name().startsWith("unsigned short")) {
+        strType = "int";
+    } else if (type->name().startsWith("unsigned ")) { // uint and ulong
+        strType = "long";
+    } else if (type->isContainer()) {
+        QString strType = translateType(type, cppClass, Options(ExcludeConst) | ExcludeReference);
+        strType.remove("*");
         strType.remove(">");
         strType.remove("<");
-
-        //replace ::
         strType.replace("::", ".");
-
-        //Translate ContainerType
         if (strType.contains("QList") || strType.contains("QVector")) {
-            strType.replace("QList", "List of ");
-            strType.replace("QVector", "List of ");
+            strType.replace("QList", "list of ");
+            strType.replace("QVector", "list of ");
         } else if (strType.contains("QHash") || strType.contains("QMap")) {
             strType.remove("QHash");
             strType.remove("QMap");
             QStringList types = strType.split(",");
             strType = QString("Dictionary with keys of type %1 and values of type %2.")
-                .arg(types[0]).arg(types[1]);
+            .arg(types[0]).arg(types[1]);
         }
-        return strType;
+    } else {
+        QString refTag;
+        if (type->isEnum())
+            refTag = "attr";
+        else
+            refTag = "class";
+        strType = ':' + refTag + ":`" + type->fullName() + '`';
     }
+    return strType;
 }
 
 void QtDocGenerator::writeParamerteType(QTextStream &s, const AbstractMetaClass *cppClass, const AbstractMetaArgument *arg)
 {
-    s << INDENT << ":param " << arg->argumentName() << ": "
+    s << INDENT << ":param " << arg->name() << ": "
       << translateToPythonType(arg->type(), cppClass) << endl;
 }
 
@@ -1293,7 +1328,21 @@ void QtDocGenerator::writeFunctionParametersType(QTextStream &s, const AbstractM
     }
 
     if (!func->isConstructor() && func->type()) {
-        s << INDENT << ":rtype: " << translateToPythonType(func->type(), cppClass) << endl;
+
+        QString retType;
+        // check if the return type was modified
+        foreach (FunctionModification mod, func->modifications()) {
+            foreach (ArgumentModification argMod, mod.argument_mods) {
+                if (argMod.index == 0) {
+                    retType = argMod.modified_type;
+                    break;
+                }
+            }
+        }
+
+        if (retType.isEmpty())
+            retType = translateToPythonType(func->type(), cppClass);
+        s << INDENT << ":rtype: " << retType << endl;
     }
     s << endl;
 }
@@ -1308,7 +1357,8 @@ void QtDocGenerator::writeFunction(QTextStream &s, bool writeDoc, const Abstract
         writeFunctionParametersType(s, cppClass, func);
         s << endl;
         writeInjectDocumentation(s, DocModification::Prepend, cppClass, func);
-        writeFormatedText(s, func->documentation(), cppClass);
+        if (!writeInjectDocumentation(s, DocModification::Replace, cppClass, func))
+            writeFormatedText(s, func->documentation(), cppClass);
         writeInjectDocumentation(s, DocModification::Append, cppClass, func);
     }
 }
@@ -1318,67 +1368,90 @@ void QtDocGenerator::finishGeneration()
     if (classes().isEmpty())
         return;
 
-    QFile input(outputDirectory() + '/' + subDirectoryForPackage(packageName()) + "/index.rst");
-    input.open(QIODevice::WriteOnly);
-    QTextStream s(&input);
+    QMap<QString, QStringList>::iterator it = m_packages.begin();
+    for (; it != m_packages.end(); ++it) {
+        QString outputDir = outputDirectory() + '/' + QString(it.key()).replace(".", "/");
+        FileOut output(outputDir + "/index.rst");
+        QTextStream& s = output.stream;
 
-    s << ".. module:: " << packageName() << endl << endl;
+        s << ".. module:: " << it.key() << endl << endl;
 
-    QString title = packageName() + " contents";
-    s << title << endl;
-    s << createRepeatedChar(title.length(), '*') << endl << endl;
-    s << ".. toctree::" << endl;
+        QString title = it.key() + " contents";
+        s << title << endl;
+        s << createRepeatedChar(title.length(), '*') << endl << endl;
+        s << ".. toctree::" << endl;
 
-    /* Avoid showing "Detailed Description for *every* class in toc tree */
-    Indentation indentation(INDENT);
-    s << INDENT << ":maxdepth: 1" << endl << endl;
+        /* Avoid showing "Detailed Description for *every* class in toc tree */
+        Indentation indentation(INDENT);
+        s << INDENT << ":maxdepth: 1" << endl << endl;
 
-    QStringList classList;
-    foreach (AbstractMetaClass *cls, classes()) {
-        if (!shouldGenerate(cls))
-            continue;
-        classList << getClassName(cls);
-    }
-    classList.sort();
+        // Search for extra-sections
+        if (!m_extraSectionDir.isEmpty()) {
+            QDir extraSectionDir(m_extraSectionDir);
+            QStringList fileList = extraSectionDir.entryList(QStringList() << (it.key() + "?*.rst"), QDir::Files);
+            QStringList::iterator it2 = fileList.begin();
+            for (; it2 != fileList.end(); ++it2) {
+                QString origFileName(*it2);
+                it2->remove(0, it.key().count() + 1);
+                QString newFilePath = outputDir + '/' + *it2;
+                if (QFile::exists(newFilePath))
+                    QFile::remove(newFilePath);
+                if (!QFile::copy(m_extraSectionDir + '/' + origFileName, newFilePath)) {
+                    ReportHandler::warning("Error copying extra doc " + (m_extraSectionDir + '/' + origFileName)
+                                           + " to " + newFilePath);
+                }
+            }
+            it.value().append(fileList);
+        }
 
-    foreach (QString clazz, classList)
-        s << INDENT << clazz << endl;
+        qSort(it.value());
+        foreach (QString className, it.value()) {
+            s << INDENT << className << endl;
+        }
+        s << endl << endl;
 
-    s << endl << endl;
+        s << "Detailed Description" << endl;
+        s << "--------------------" << endl << endl;
 
-    s << "Detailed Description" << endl;
-    s << "--------------------" << endl << endl;
-
-    if (m_moduleDoc.format() == Documentation::Native) {
-        QtXmlToSphinx x(this, m_moduleDoc.value(), moduleName());
-        s << x;
-    } else {
-        s << m_moduleDoc.value();
+        // module doc is always wrong and C++istic, so go straight to the extra directory!
+        QFile moduleDoc(m_extraSectionDir + '/' + it.key() + ".rst");
+        if (moduleDoc.open(QIODevice::ReadOnly | QIODevice::Text)) {
+            s << moduleDoc.readAll();
+            moduleDoc.close();
+        } else {
+            // try the normal way
+            Documentation moduleDoc = m_docParser->retrieveModuleDocumentation(it.key());
+            if (moduleDoc.format() == Documentation::Native) {
+                QtXmlToSphinx x(this, moduleDoc.value(), QString(it.key()).remove(0, it.key().lastIndexOf('.') + 1));
+                s << x;
+            } else {
+                s << moduleDoc.value();
+            }
+        }
     }
 }
 
 bool QtDocGenerator::doSetup(const QMap<QString, QString>& args)
 {
     m_libSourceDir = args.value("library-source-dir");
-    setOutputDirectory(args.value("documentation-out-dir"));
     m_docDataDir = args.value("documentation-data-dir");
-    m_codeSnippetDir = args.value("documentation-code-snippets-dir", m_libSourceDir);
+#ifdef __WIN32__
+#   define PATH_SEP ";"
+#else
+#   define PATH_SEP ":"
+#endif
+    m_codeSnippetDirs = args.value("documentation-code-snippets-dir", m_libSourceDir).split(PATH_SEP);
+    m_extraSectionDir = args.value("documentation-extra-sections-dir");
 
     if (m_libSourceDir.isEmpty() || m_docDataDir.isEmpty()) {
         ReportHandler::warning("Documentation data dir and/or Qt source dir not informed, "
                                "documentation will not be extracted from Qt sources.");
         return false;
     } else {
-        QtDocParser docParser;
-        docParser.setPackageName(packageName());
-        docParser.setDocumentationDataDirectory(m_docDataDir);
-        docParser.setLibrarySourceDirectory(m_libSourceDir);
-        foreach(AbstractMetaClass* cppClass, classes()) {
-            docParser.fillDocumentation(cppClass);
-        }
-        m_moduleDoc = docParser.retrieveModuleDocumentation();
-        return true;
+        m_docParser->setDocumentationDataDirectory(m_docDataDir);
+        m_docParser->setLibrarySourceDirectory(m_libSourceDir);
     }
+    return true;
 }
 
 
@@ -1386,9 +1459,9 @@ QMap<QString, QString> QtDocGenerator::options() const
 {
     QMap<QString, QString> options;
     options.insert("library-source-dir", "Directory where library source code is located");
-    options.insert("documentation-out-dir", "The directory where the generated documentation files will be written");
     options.insert("documentation-data-dir", "Directory with XML files generated by documentation tool (qdoc3 or Doxygen)");
     options.insert("documentation-code-snippets-dir", "Directory used to search code snippets used by the documentation");
+    options.insert("documentation-extra-sections-dir", "Directory used to search for extra documentation sections");
     return options;
 }
 
